# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
from .zlib_base64_to_str import (
    zlib_to_dict,
    zlib_to_str,
)
from .str_to_zlib_base64 import (
    dict_to_zlib,
    str_to_zlib,
)



# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
from qiskit.result import Result


def _hex_to_bitstring(value, nb_bits):
    mask = (1 << nb_bits) - 1
    return format(value & mask, f"0{nb_bits}b")


def convert(result_dict: dict, **kwargs) -> Result:
    results = result_dict["results"]

    for experiment in results:
        exp_data = experiment.get("data", {})
        counts = exp_data.get("counts", None)
        n_qubits = experiment["header"]["n_qubits"]

        if counts:
            new_counts = {}

            for bitstring, count in counts.items():
                if bitstring.startswith("0x"):
                    nbits = _hex_to_bitstring(int(bitstring, 16), n_qubits)
                else:
                    nbits = bitstring

                new_counts[nbits] = count

            exp_data["counts"] = new_counts

    data = {
        "results": results,
        "success": result_dict["success"],
        "header": result_dict.get("header"),
        "metadata": result_dict.get("metadata"),
    }

    if kwargs:
        data.update(kwargs)

    return Result.from_dict(data)



# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
import numpy as np

from cirq import ResultDict


def _extract_measurement_key(experiment_result: dict) -> str:
    header = experiment_result.get("header", {})
    creg_sizes = header.get("creg_sizes", [])

    if creg_sizes and len(creg_sizes) > 0:
        m_name = creg_sizes[0][0]
        return m_name.replace("m_", "")

    return "m"


def convert(result_dict: dict) -> ResultDict:
    experiment = result_dict.get("results", [{}])[0]
    m_key = _extract_measurement_key(experiment)

    counts = experiment.get("data", {}).get("counts", {})
    header = experiment.get("header", {})
    qreg_sizes = header.get("qreg_sizes", [])
    num_qubits = header.get("n_qubits", None)

    if not num_qubits and qreg_sizes and len(qreg_sizes) > 0:
        num_qubits = qreg_sizes[0][1]
    else:
        memory = experiment.get("memory", {})

        if memory and len(memory) > 0:
            num_qubits = len(memory[0])

    all_shots = []

    for bitstring_hex, count in counts.items():
        if bitstring_hex.startswith("0x"):
            integer_val = int(bitstring_hex, 16)
            bitstring = format(integer_val, f"0{num_qubits}b")
        else:
            bitstring = bitstring_hex

        bits = [int(b) for b in bitstring]
        for _ in range(count):
            all_shots.append(bits)

    measurements = {m_key: np.array(all_shots)}

    return ResultDict(params=result_dict.pop("params", None), measurements=measurements)



import cudaq
import numpy as np

from cudaq import spin

from qio.core import (
    QuantumProgramResult,
    QuantumProgramResultCompressionFormat,
    QuantumProgram,
)


def test_global_cudaq_flow():
    ### Server side
    shots = 1000
    qubit_count = 4

    @cudaq.kernel
    def dummy_kernel(qubit_count: int):
        qubits = cudaq.qvector(qubit_count)
        h(qubits[0])
        for i in range(1, qubit_count):
            x.ctrl(qubits[0], qubits[i])
        mz(qubits)

    results = cudaq.sample(dummy_kernel, qubit_count, shots_count=shots)

    qpr_json = QuantumProgramResult.from_cudaq_sample_result(
        results, compression_format=QuantumProgramResultCompressionFormat.NONE
    ).to_json_str()

    compressed_qpr_json = QuantumProgramResult.from_cudaq_sample_result(
        results, compression_format=QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1
    ).to_json_str()

    assert qpr_json is not None
    assert compressed_qpr_json is not None

    ### Client side
    qpr = QuantumProgramResult.from_json_str(qpr_json)
    compressed_qpr = QuantumProgramResult.from_json_str(compressed_qpr_json)

    cudaq_result = qpr.to_cudaq_sample_result()
    assert cudaq_result is not None
    print("cudaq result:", cudaq_result)

    uncomp_cudaq_result = compressed_qpr.to_cudaq_sample_result()
    assert uncomp_cudaq_result is not None
    print("cudaq result from compressed data:", uncomp_cudaq_result)

    assert cudaq_result.get_total_shots() == uncomp_cudaq_result.get_total_shots()
    assert cudaq_result.most_probable() == uncomp_cudaq_result.most_probable()
    assert cudaq_result.serialize() == uncomp_cudaq_result.serialize()

    qiskit_result = compressed_qpr.to_qiskit_result()
    print("cudaq result as qiskit result:", qiskit_result)
    assert qiskit_result is not None


def test_simple_kernel():

    @cudaq.kernel
    def kernel():
        q = cudaq.qubit()
        x(q)
        mz(q)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()

    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert len(counts) == 1
    assert "1" in counts


def test_multi_qubit_kernel():

    @cudaq.kernel
    def kernel():
        q0 = cudaq.qubit()
        q1 = cudaq.qubit()
        h(q0)
        cx(q0, q1)
        mz(q0)
        mz(q1)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()

    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert len(counts) == 2
    assert "00" in counts
    assert "11" in counts


def test_qvector_kernel():

    @cudaq.kernel
    def kernel():
        qubits = cudaq.qvector(2)
        h(qubits[0])
        x.ctrl(qubits[0], qubits[1])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()

    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert len(counts) == 2
    assert "00" in counts
    assert "11" in counts


def test_builder_sample():

    kernel = cudaq.make_kernel()
    qubits = kernel.qalloc(2)
    kernel.h(qubits[0])
    kernel.cx(qubits[0], qubits[1])
    kernel.mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()

    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert len(counts) == 2
    assert "00" in counts
    assert "11" in counts


def test_all_gates():

    @cudaq.kernel
    def single_qubit_gates():
        q = cudaq.qubit()
        h(q)
        x(q)
        y(q)
        z(q)
        r1(np.pi, q)
        rx(np.pi, q)
        ry(np.pi, q)
        rz(np.pi, q)
        s(q)
        t(q)
        mx(q)
        ## my(q) # not supported since the default rewriter uses `sdg`
        mz(q)

    # Test here is that this runs
    program = QuantumProgram.from_cudaq_kernel(single_qubit_gates)
    kernel_from_program = program.to_cudaq_kernel()

    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert counts

    @cudaq.kernel
    def two_qubit_gates():
        qubits = cudaq.qvector(2)
        x(qubits[0])
        swap(qubits[0], qubits[1])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(two_qubit_gates)
    kernel_from_program = program.to_cudaq_kernel()

    counts = cudaq.sample(kernel_from_program, shots_count=100)

    counts = cudaq.sample(two_qubit_gates, shots_count=100)

    assert len(counts) == 1
    assert "01" in counts


# def test_multi_qvector():

#     @cudaq.kernel
#     def kernel():
#         qubits = cudaq.qvector(2)
#         ancilla = cudaq.qvector(2)
#         x(qubits)
#         h(ancilla)
#         mz(ancilla)

#     program = QuantumProgram.from_cudaq_kernel(kernel)
#     kernel_from_program = program.to_cudaq_kernel()

#     counts = cudaq.sample(kernel_from_program, shots_count=100)

#     assert counts


def test_control_modifier():

    @cudaq.kernel
    def single_qubit_gates():
        qubits = cudaq.qvector(2)
        h.ctrl(qubits[0], qubits[1])
        x.ctrl(qubits[1], qubits[0])
        y.ctrl(qubits[0], qubits[1])
        z.ctrl(qubits[1], qubits[0])
        r1.ctrl(np.pi / 2, qubits[0], qubits[1])
        rx.ctrl(np.pi / 4, qubits[1], qubits[0])
        ry.ctrl(np.pi / 8, qubits[0], qubits[1])
        rz.ctrl(np.pi, qubits[1], qubits[0])
        s.ctrl(qubits[0], qubits[1])
        t.ctrl(qubits[1], qubits[0])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(single_qubit_gates)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)

    @cudaq.kernel
    def two_qubit_gates():
        qubits = cudaq.qvector(3)
        x(qubits[0])
        x(qubits[1])
        swap.ctrl(qubits[0], qubits[1], qubits[2])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(two_qubit_gates)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert len(counts) == 1
    assert "101" in counts

    @cudaq.kernel
    def bell():
        qubits = cudaq.qvector(2)
        h(qubits[0])
        x.ctrl(qubits[0], qubits[1])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(bell)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert len(counts) == 2
    assert "00" in counts
    assert "11" in counts


def test_adjoint_modifier():

    @cudaq.kernel
    def single_qubit_gates():
        q = cudaq.qubit()
        r1.adj(np.pi, q)
        rx.adj(np.pi / 2, q)
        ry.adj(np.pi / 4, q)
        rz.adj(np.pi / 8, q)
        s.adj(q)
        t.adj(q)
        mz(q)

    # Test here is that this runs
    program = QuantumProgram.from_cudaq_kernel(single_qubit_gates)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)

    assert counts


def test_u3_decomposition():

    @cudaq.kernel
    def kernel():
        qubit = cudaq.qubit()
        u3(0.0, np.pi / 2, np.pi, qubit)
        mz(qubit)

    # Test here is that this runs
    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    result = cudaq.sample(kernel_from_program, shots_count=100)
    measurement_probabilities = dict(result.items())
    print(measurement_probabilities)

    @cudaq.kernel
    def kernel():
        qubits = cudaq.qvector(2)
        u3.ctrl(0.0, np.pi / 2, np.pi, qubits[0], qubits[1])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)
    print("test_u3_decomposition bell", counts)

    assert "00" in counts
    assert len(counts) == 1


def test_sample_async():
    cudaq.set_random_seed(13)

    @cudaq.kernel
    def simple():
        qubits = cudaq.qvector(2)
        h(qubits[0])
        x.ctrl(qubits[0], qubits[1])
        mz(qubits)

    program = QuantumProgram.from_cudaq_kernel(simple)
    kernel_from_program = program.to_cudaq_kernel()
    future = cudaq.sample_async(kernel_from_program, shots_count=100)
    counts = future.get()
    print("test_sample_async bell", counts)

    assert len(counts) == 2
    assert "00" in counts
    assert "11" in counts


def test_observe_async():

    @cudaq.kernel
    def kernel():
        qubits = cudaq.qvector(2)
        x(qubits[0])

    hamiltonian = spin.z(0) * spin.z(1)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    future = cudaq.observe_async(kernel_from_program, hamiltonian, shots_count=1)
    result = future.get()

    assert result.expectation() == -1.0


def test_custom_operations():
    cudaq.register_operation("custom_x", np.array([0, 1, 1, 0]))

    @cudaq.kernel
    def basic_x():
        qubit = cudaq.qubit()
        custom_x(qubit)
        mz(qubit)

    program = QuantumProgram.from_cudaq_kernel(basic_x)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)
    print("test_custom_operations", counts)

    assert len(counts) == 1 and "1" in counts


# def test_kernel_with_args():

#     @cudaq.kernel
#     def kernel(qubit_count: int):
#         qreg = cudaq.qvector(qubit_count)
#         h(qreg[0])
#         for qubit in range(qubit_count - 1):
#             x.ctrl(qreg[qubit], qreg[qubit + 1])
#         mz(qreg)

#     program = QuantumProgram.from_cudaq_kernel(kernel)
#     kernel_from_program = program.to_cudaq_kernel()
#     counts = cudaq.sample(kernel_from_program, 4, shots_count=100)
#     print("test_kernel_with_args", counts)

#     assert len(counts) == 2
#     assert "0000" in counts
#     assert "1111" in counts


# def test_kernel_subveqs():

#     @cudaq.kernel
#     def kernel():
#         qreg = cudaq.qvector(4)
#         x(qreg[1])
#         x(qreg[2])
#         v = qreg[1:3]
#         mz(v)

#     program = QuantumProgram.from_cudaq_kernel(kernel)
#     kernel_from_program = program.to_cudaq_kernel()
#     counts = cudaq.sample(kernel_from_program, shots_count=100)
#     print("test_kernel_subveqs", counts)

#     assert len(counts) == 1
#     assert "11" in counts  # got 0110


# def test_kernel_two_subveqs():

#     @cudaq.kernel
#     def kernel():
#         qreg = cudaq.qvector(4)
#         x(qreg[1])
#         x(qreg[2])
#         v1 = qreg[0:2]
#         mz(v1)
#         v2 = qreg[2:3]
#         mz(v2)

#     program = QuantumProgram.from_cudaq_kernel(kernel)
#     kernel_from_program = program.to_cudaq_kernel()
#     counts = cudaq.sample(kernel_from_program, shots_count=100)
#     print("test_kernel_two_subveqs", counts)

#     assert len(counts) == 1
#     assert "011" in counts  # got 0110


# def test_kernel_qubit_subveq():

#     @cudaq.kernel
#     def kernel():
#         qreg = cudaq.qvector(4)
#         x(qreg[1])
#         x(qreg[2])
#         v1 = qreg[0:2]
#         mz(v1)
#         v2 = qreg[2]
#         mz(v2)

#     program = QuantumProgram.from_cudaq_kernel(kernel)
#     kernel_from_program = program.to_cudaq_kernel()
#     counts = cudaq.sample(kernel_from_program, shots_count=100)
#     print("test_kernel_qubit_subveq", counts)

#     assert len(counts) == 1
#     assert "011" in counts  # got 0110


def test_multiple_measurement():

    @cudaq.kernel
    def kernel():
        qubits = cudaq.qvector(2)
        h(qubits[0])
        mz(qubits[0])
        mz(qubits[1])

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)
    print("test_multiple_measurement", counts)

    assert len(counts) == 2
    assert "00" in counts
    assert "10" in counts


def test_multiple_measurement_non_consecutive():

    @cudaq.kernel
    def kernel():
        qubits = cudaq.qvector(3)
        x(qubits[0])
        x(qubits[2])
        mz(qubits[0])
        mz(qubits[2])

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)
    print("test_multiple_measurement_non_consecutive", counts)

    assert len(counts) == 1
    assert "11" in counts


def test_qvector_slicing():

    @cudaq.kernel
    def kernel():
        q = cudaq.qvector(4)
        x(q.front(2))
        mz(q)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)
    print("test_qvector_slicing", counts)

    assert len(counts) == 1
    assert "1100" in counts


# def test_toffoli():

#     @cudaq.kernel
#     def kernel():
#         q = cudaq.qvector(3)
#         x(q)
#         x.ctrl([q[0], q[1]], q[2])
#         mz(q)

#     program = QuantumProgram.from_cudaq_kernel(kernel)
#     kernel_from_program = program.to_cudaq_kernel()
#     counts = cudaq.sample(kernel_from_program, shots_count=100)
#     counts.dump()


def test_state_prep():

    @cudaq.kernel
    def kernel():
        q = cudaq.qvector([1.0 / np.sqrt(2.0), 0.0, 0.0, 1.0 / np.sqrt(2.0)])
        mz(q)

    program = QuantumProgram.from_cudaq_kernel(kernel)
    kernel_from_program = program.to_cudaq_kernel()
    counts = cudaq.sample(kernel_from_program, shots_count=100)
    counts.dump()

    assert "11" in counts
    assert "00" in counts



# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from typing import Dict, List, Tuple


def _long_to_bitstring(val: int, size: int) -> str:
    """
    Equivalent to the C++ longToBitString.
    Converts an integer to a binary string padded to the correct bit length.
    """
    # Format to binary, remove '0b' prefix, and pad with leading zeros
    return bin(val)[2:].zfill(size)


def _extract_name(data: List[int], stride: int) -> Tuple[str, int]:
    """Extracts a string (register name) from the data array."""
    n_chars = data[stride]
    stride += 1
    name = "".join(chr(data[i]) for i in range(stride, stride + n_chars))
    stride += n_chars

    return name, stride


def _deserialize_to_dict(data: List[int]) -> Dict[str, Dict[str, int]]:
    """
    Parses the integer array into a dictionary of registers and their counts.
    Matches the logic of ExecutionResult::deserialize and deserializeCounts.
    """
    stride = 0
    all_results = {}

    while stride < len(data):
        # 1. Extract Register Name
        name, stride = _extract_name(data, stride)

        # 2. Extract Counts (deserializeCounts logic)
        num_bitstrings = data[stride]
        stride += 1

        local_counts = {}
        memory = []
        # Each entry is a triplet: [packed_value, bit_size, count]
        for _ in range(num_bitstrings):
            bitstring_as_long = data[stride]
            size_of_bitstring = data[stride + 1]
            count = data[stride + 2]

            bs = _long_to_bitstring(bitstring_as_long, size_of_bitstring)
            local_counts[bs] = count
            memory.extend([bs] * count)
            stride += 3

        all_results[name] = local_counts

    return all_results, memory


def convert(result_str: str, **kwargs) -> Result:
    parsed_data, memory = _deserialize_to_dict(result_str)
    experiment_results = []

    for reg_name, counts in parsed_data.items():
        shots = sum(counts.values())

        # Encapsulate data in Qiskit's expected format
        data_payload = ExperimentResultData(counts=counts, memory=memory)

        exp_res = ExperimentResult(
            shots=shots,
            success=True,
            data=data_payload,
            header={"name": reg_name, "memory": True},
            status="Done",
        )
        experiment_results.append(exp_res)

    kwargs = kwargs or {}
    return Result(results=experiment_results, **kwargs)



# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
import json

from enum import IntEnum
from typing import Dict

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from qio.utils.compression import zlib_to_str, str_to_zlib


class QuantumProgramSerializationFormat(IntEnum):
    UNKNOWN_SERIALIZATION_FORMAT = 0
    QASM_V1 = 1
    QASM_V2 = 2
    QASM_V3 = 3
    QIR_V1 = 4
    CIRQ_CIRCUIT_JSON_V1 = 5
    PERCEVAL_CIRCUIT_JSON_V1 = 6
    PULSER_SEQUENCE_JSON_V1 = 7


class QuantumProgramCompressionFormat(IntEnum):
    UNKNOWN_COMPRESSION_FORMAT = 0
    NONE = 1
    ZLIB_BASE64_V1 = 2


@dataclass_json
@dataclass
class QuantumProgram:
    compression_format: QuantumProgramCompressionFormat
    serialization_format: QuantumProgramSerializationFormat
    serialization: str

    @classmethod
    def from_json_dict(cls, data: Dict) -> "QuantumProgram":
        return QuantumProgram.from_dict(data)

    def to_json_dict(self) -> Dict:
        return self.to_dict()

    @classmethod
    def from_json_str(cls, data: str) -> "QuantumProgram":
        while isinstance(data, str):
            data = json.loads(data)
        return cls.from_json_dict(data)

    def to_json_str(self) -> str:
        return json.dumps(self.to_json_dict())

    @classmethod
    def from_qiskit_circuit(
        cls,
        qiskit_circuit: "qiskit.QuantumCircuit",
        dest_format: QuantumProgramSerializationFormat = QuantumProgramSerializationFormat.QASM_V3,
        compression_format: QuantumProgramCompressionFormat = QuantumProgramCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgram":
        from qio.utils.conversion.program.qiskit_to_qasm2 import (
            convert as qiskit_to_qasm2_convert,
        )
        from qio.utils.conversion.program.qiskit_to_qasm3 import (
            convert as qiskit_to_qasm3_convert,
        )

        dest_format = (
            QuantumProgramSerializationFormat.QASM_V3
            if dest_format
            == QuantumProgramSerializationFormat.UNKNOWN_SERIALIZATION_FORMAT
            else dest_format
        )
        compression_format = (
            QuantumProgramCompressionFormat.NONE
            if compression_format
            == QuantumProgramCompressionFormat.UNKNOWN_COMPRESSION_FORMAT
            else compression_format
        )

        try:
            apply_serialization = {
                QuantumProgramSerializationFormat.QASM_V2: qiskit_to_qasm2_convert,
                QuantumProgramSerializationFormat.QASM_V3: qiskit_to_qasm3_convert,
            }

            serialization = apply_serialization[dest_format](qiskit_circuit)

            if compression_format == QuantumProgramCompressionFormat.ZLIB_BASE64_V1:
                serialization = str_to_zlib(serialization)

            return cls(
                serialization_format=dest_format,
                compression_format=compression_format,
                serialization=serialization,
            )
        except Exception as e:
            raise Exception(
                "unsupport serialization:", dest_format, compression_format, e
            )

    def to_qiskit_circuit(self) -> "qiskit.QuantumCircuit":
        from qio.utils.conversion.program.qasm1_to_qiskit import (
            convert as qasm1_to_qiskit_convert,
        )
        from qio.utils.conversion.program.qasm2_to_qiskit import (
            convert as qasm2_to_qiskit_convert,
        )
        from qio.utils.conversion.program.qasm3_to_qiskit import (
            convert as qasm3_to_qiskit_convert,
        )

        serialization = self.serialization

        if self.compression_format == QuantumProgramCompressionFormat.ZLIB_BASE64_V1:
            serialization = zlib_to_str(serialization)

        try:
            apply_unserialization = {
                QuantumProgramSerializationFormat.QASM_V1: qasm1_to_qiskit_convert,
                QuantumProgramSerializationFormat.QASM_V2: qasm2_to_qiskit_convert,
                QuantumProgramSerializationFormat.QASM_V3: qasm3_to_qiskit_convert,
            }

            return apply_unserialization[self.serialization_format](serialization)
        except Exception as e:
            raise Exception(
                "unsupported unserialization:", self.serialization_format, e
            )

    @classmethod
    def from_cirq_circuit(
        cls,
        cirq_circuit: "cirq.AbstractCircuit",
        dest_format: QuantumProgramSerializationFormat = QuantumProgramSerializationFormat.CIRQ_CIRCUIT_JSON_V1,
        compression_format: QuantumProgramCompressionFormat = QuantumProgramCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgram":
        from qio.utils.conversion.program.cirq_to_cirqjson import (
            convert as cirq_to_cirqjson_convert,
        )
        from qio.utils.conversion.program.cirq_to_qasm2 import (
            convert as cirq_to_qasm2_convert,
        )
        from qio.utils.conversion.program.cirq_to_qasm3 import (
            convert as cirq_to_qasm3_convert,
        )

        dest_format = (
            QuantumProgramSerializationFormat.CIRQ_CIRCUIT_JSON_V1
            if dest_format
            == QuantumProgramSerializationFormat.UNKNOWN_SERIALIZATION_FORMAT
            else dest_format
        )
        compression_format = (
            QuantumProgramCompressionFormat.NONE
            if compression_format
            == QuantumProgramCompressionFormat.UNKNOWN_COMPRESSION_FORMAT
            else compression_format
        )

        try:
            apply_serialization = {
                QuantumProgramSerializationFormat.QASM_V2: cirq_to_qasm2_convert,
                QuantumProgramSerializationFormat.QASM_V3: cirq_to_qasm3_convert,
                QuantumProgramSerializationFormat.CIRQ_CIRCUIT_JSON_V1: cirq_to_cirqjson_convert,
            }

            serialization = apply_serialization[dest_format](cirq_circuit)

            if compression_format == QuantumProgramCompressionFormat.ZLIB_BASE64_V1:
                serialization = str_to_zlib(serialization)

            return cls(
                serialization_format=dest_format,
                compression_format=compression_format,
                serialization=serialization,
            )
        except Exception as e:
            raise Exception("unsupported serialization:", dest_format, e)

    def to_cirq_circuit(self) -> "cirq.Circuit":
        from qio.utils.conversion.program.qasm1_to_cirq import (
            convert as qasm1_to_cirq_convert,
        )
        from qio.utils.conversion.program.qasm2_to_cirq import (
            convert as qasm2_to_cirq_convert,
        )
        from qio.utils.conversion.program.qasm3_to_cirq import (
            convert as qasm3_to_cirq_convert,
        )
        from qio.utils.conversion.program.cirqjson_to_cirq import (
            convert as cirqjson_to_cirq_convert,
        )

        serialization = self.serialization

        try:
            if (
                self.compression_format
                == QuantumProgramCompressionFormat.ZLIB_BASE64_V1
            ):
                serialization = zlib_to_str(serialization)

            apply_serialization = {
                QuantumProgramSerializationFormat.QASM_V1: qasm1_to_cirq_convert,
                QuantumProgramSerializationFormat.QASM_V2: qasm2_to_cirq_convert,
                QuantumProgramSerializationFormat.QASM_V3: qasm3_to_cirq_convert,
                QuantumProgramSerializationFormat.CIRQ_CIRCUIT_JSON_V1: cirqjson_to_cirq_convert,
            }

            return apply_serialization[self.serialization_format](serialization)

        except Exception as e:
            raise Exception(
                "unsupported unserialization:", self.serialization_format, e
            )

    @classmethod
    def from_cudaq_kernel(
        cls,
        cudaq_kernel: "cudaq.PyKernel",
        dest_format: QuantumProgramSerializationFormat = QuantumProgramSerializationFormat.QASM_V2,
        compression_format: QuantumProgramCompressionFormat = QuantumProgramCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgram":
        from qio.utils.conversion.program.cudaq_to_qasm2 import (
            convert as cudaq_to_qasm2_convert,
        )

        dest_format = (
            QuantumProgramSerializationFormat.QASM_V2
            if dest_format
            == QuantumProgramSerializationFormat.UNKNOWN_SERIALIZATION_FORMAT
            else dest_format
        )
        compression_format = (
            QuantumProgramCompressionFormat.NONE
            if compression_format
            == QuantumProgramCompressionFormat.UNKNOWN_COMPRESSION_FORMAT
            else compression_format
        )

        try:
            apply_serialization = {
                QuantumProgramSerializationFormat.QASM_V2: cudaq_to_qasm2_convert,
            }

            serialization = apply_serialization[dest_format](cudaq_kernel)

            if compression_format == QuantumProgramCompressionFormat.ZLIB_BASE64_V1:
                serialization = str_to_zlib(serialization)

            return cls(
                serialization_format=dest_format,
                compression_format=compression_format,
                serialization=serialization,
            )
        except Exception as e:
            raise Exception("unsupported serialization:", dest_format, e)

    def to_cudaq_kernel(self) -> "cudaq.Kernel":
        from qio.utils.conversion.program.qasm2_to_cudaq import (
            convert as qasm2_to_cudaq_convert,
        )
        from qio.utils.conversion.program.qasm3_to_cudaq import (
            convert as qasm3_to_cudaq_convert,
        )

        serialization = self.serialization

        try:
            if (
                self.compression_format
                == QuantumProgramCompressionFormat.ZLIB_BASE64_V1
            ):
                serialization = zlib_to_str(serialization)

            apply_unserialization = {
                QuantumProgramSerializationFormat.QASM_V2: qasm2_to_cudaq_convert,
                QuantumProgramSerializationFormat.QASM_V3: qasm3_to_cudaq_convert,
            }

            return apply_unserialization[self.serialization_format](serialization)
        except Exception as e:
            raise Exception(
                "unsupported unserialization:", self.serialization_format, e
            )



# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from qiskit_aer import noise

from qio.core import QuantumNoiseModel


def test_aer_noise():
    prob_1 = 0.01  # 1-qubit gate
    prob_2 = 0.1  # 2-qubit gate

    # Depolarizing quantum errors
    error_1 = noise.depolarizing_error(prob_1, 1)
    error_2 = noise.depolarizing_error(prob_2, 2)

    # Add errors to noise model
    noise_model = noise.NoiseModel()
    noise_model.add_all_qubit_quantum_error(error_1, ["rz", "sx", "x"])
    noise_model.add_all_qubit_quantum_error(error_2, ["cx"])

    qnoise_model = QuantumNoiseModel.from_qiskit_aer_noise_model(noise_model)

    deser_noise_model = qnoise_model.to_qiskit_aer_noise_model()

    assert deser_noise_model == noise_model



# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
import json
import zlib
import base64

from typing import Dict


def zlib_to_str(e: str) -> str:
    base64_payload = e.encode("ascii")
    compressed_payload = base64.b64decode(base64_payload)
    json_bytes_payload = zlib.decompress(compressed_payload)
    string_payload = json_bytes_payload.decode()

    return string_payload


def zlib_to_dict(e: str) -> Dict:
    s = zlib_to_str(e)
    dict = json.loads(s)

    return dict



# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .quantum_program import (
    QuantumProgram,
    QuantumProgramSerializationFormat,
    QuantumProgramCompressionFormat,
)
from .quantum_program_result import (
    QuantumProgramResult,
    QuantumProgramResultSerializationFormat,
    QuantumProgramResultCompressionFormat,
)
from .quantum_computation_models import (
    QuantumComputationModel,
    QuantumComputationParameters,
    BackendData,
    ClientData,
)
from .quantum_noise_model import (
    QuantumNoiseModelSerializationFormat,
    QuantumNoiseModelCompressionFormat,
    QuantumNoiseModel,
)



# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
import cirq
import re

from collections import defaultdict
from cirq.contrib.qasm_import import circuit_from_qasm


def _restore_terminal_measurements(circuit: cirq.Circuit) -> cirq.Circuit:
    groups = defaultdict(dict)
    ops_to_remove = []

    pattern = re.compile(r"^m_(.+)_(?P<idx>\d+)$")

    # 1. Identify terminal segmented measurements
    for i, moment in enumerate(circuit):
        for op in moment:
            if isinstance(op.gate, cirq.MeasurementGate):
                key = op.gate.key
                match = pattern.match(key)

                if match:
                    qubit = op.qubits[0]
                    # Check if this is the last moment in the circuit
                    if circuit.next_moment_operating_on([qubit], i + 1) is None:
                        original_name = match.group(1)
                        index = int(match.group("idx"))
                        groups[original_name][index] = qubit
                        ops_to_remove.append((i, op))

    if not groups:
        return circuit

    # 2. Cleanup circuit
    # Copy to avoid modifying original in place
    new_circuit = circuit.copy()
    new_circuit.batch_remove(ops_to_remove)

    # 3. Add merged measurements at the end
    for name in sorted(groups.keys()):
        indexed_qubits = groups[name]
        sorted_indices = sorted(indexed_qubits.keys())
        ordered_qubits = [indexed_qubits[idx] for idx in sorted_indices]
        new_circuit.append(cirq.measure(*ordered_qubits, key=name))

    return new_circuit


def convert(circuit_str: str) -> cirq.Circuit:
    return _restore_terminal_measurements(circuit_from_qasm(circuit_str))



# Copyright 2026 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

import numpy as np
import collections
import io

from typing import Union, Sequence, Dict, Tuple, Callable, TypeVar, cast

T = TypeVar("T")


def _unpack_bits(packed_bits: str, dtype: str, shape: Sequence[int]) -> np.ndarray:
    bits_bytes = bytes.fromhex(packed_bits)
    bits = np.unpackbits(np.frombuffer(bits_bytes, dtype=np.uint8))
    return bits[: np.prod(shape).item()].reshape(shape).astype(dtype)


def _unpack_digits(
    packed_digits: str,
    binary: bool,
    dtype: Union[None, str],
    shape: Union[None, Sequence[int]],
):
    if binary:
        dtype = cast(str, dtype)
        shape = cast(Sequence[int], shape)
        return _unpack_bits(packed_digits, dtype, shape)

    buffer = io.BytesIO()
    buffer.write(bytes.fromhex(packed_digits))
    buffer.seek(0)
    digits = np.load(buffer, allow_pickle=False)
    buffer.close()

    return digits


def _key_to_str(key) -> str:
    if isinstance(key, str):
        return key
    return ",".join(str(q) for q in key)


def _big_endian_bits_to_int(bits) -> int:
    result = 0
    for e in bits:
        result <<= 1
        if e:
            result |= 1
    return result


def _tuple_of_big_endian_int(bit_groups) -> Tuple[int, ...]:
    return tuple(_big_endian_bits_to_int(bits) for bits in bit_groups)


def _multi_measurement_histogram(
    keys,
    measurements,
    repetitions,
    fold_func: Callable[[Tuple], T] = cast(
        Callable[[Tuple], T], _tuple_of_big_endian_int
    ),
) -> Tuple[collections.Counter, list]:
    fixed_keys = tuple(_key_to_str(key) for key in keys)
    samples = zip(*(measurements[sub_key] for sub_key in fixed_keys))

    if len(fixed_keys) == 0:
        samples = [()] * repetitions

    counter = collections.Counter()
    memory = []

    for sample in samples:
        memory.append("".join(str(a) for a in np.concatenate(sample)))
        counter[fold_func(sample)] += 1

    return (counter, memory)


def __make_hex_from_result_array(result: Tuple):
    str_value = "".join(map(str, result))
    binary_value = bin(int(str_value))
    integer_value = int(binary_value, 2)

    return hex(integer_value)


def __make_bin_from_result_array(result: Tuple):
    str_value = "".join(map(str, result))
    return str_value


def __measurements(records: Dict):
    measurements = {}
    for key, data in records.items():
        reps, instances, qubits = data.shape
        if instances != 1:
            raise ValueError("Cannot extract 2D measurements for repeated keys")
        measurements[key] = data.reshape((reps, qubits))

    return measurements


def __make_expresult_from_cirq_result(
    cirq_result_dict: Dict,
) -> ExperimentResult:
    raw_records = cirq_result_dict["records"]
    records = {key: _unpack_digits(**val) for key, val in raw_records.items()}
    measurements = __measurements(records)
    repetitions = len(next(iter(records.values())))

    counter, memory = _multi_measurement_histogram(
        keys=measurements.keys(),
        measurements=measurements,
        repetitions=repetitions,
    )

    histogram = dict(counter)

    return ExperimentResult(
        shots=repetitions,
        success=True,
        data=ExperimentResultData(
            counts={
                __make_bin_from_result_array(key): value
                for key, value in histogram.items()
            },
            memory=memory,
        ),
    )


def convert(circuit_str: str, **kwargs) -> Result:
    kwargs = kwargs or {}

    return Result(results=[__make_expresult_from_cirq_result(circuit_str)], **kwargs)



# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from typing import Union, Dict
from enum import IntEnum

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from qio.utils.compression import dict_to_zlib, zlib_to_dict


class QuantumProgramResultSerializationFormat(IntEnum):
    UNKNOWN_SERIALIZATION_FORMAT = 0
    CIRQ_RESULT_JSON_V1 = 1
    QISKIT_RESULT_JSON_V1 = 2
    CUDAQ_SAMPLE_RESULT_JSON_V1 = 3


class QuantumProgramResultCompressionFormat(IntEnum):
    UNKNOWN_COMPRESSION_FORMAT = 0
    NONE = 1
    ZLIB_BASE64_V1 = 2


@dataclass_json
@dataclass
class QuantumProgramResult:
    compression_format: QuantumProgramResultCompressionFormat
    serialization_format: QuantumProgramResultSerializationFormat
    serialization: str

    @classmethod
    def from_json_dict(cls, data: Dict) -> "QuantumProgramResult":
        return QuantumProgramResult.from_dict(data)

    def to_json_dict(self) -> Dict:
        return self.to_dict()

    @classmethod
    def from_json_str(cls, data: str) -> "QuantumProgramResult":
        while isinstance(data, str):
            data = json.loads(data)
        return cls.from_json_dict(data)

    def to_json_str(self) -> str:
        return json.dumps(self.to_json_dict())

    @classmethod
    def from_cudaq_sample_result(
        cls,
        sample_result: "cudaq.SampleResult",
        compression_format: QuantumProgramResultCompressionFormat = QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgramResult":
        from qio.utils.conversion.program_result.cudaq_sample_to_dict import (
            convert as cudaq_sample_to_dict,
        )

        compression_format = (
            QuantumProgramResultCompressionFormat.NONE
            if compression_format
            == QuantumProgramResultCompressionFormat.UNKNOWN_COMPRESSION_FORMAT
            else compression_format
        )

        apply_compression = {
            QuantumProgramResultCompressionFormat.NONE: json.dumps,
            QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1: dict_to_zlib,
        }

        try:
            sample_serialization = cudaq_sample_to_dict(sample_result)

            return cls(
                compression_format=compression_format,
                serialization_format=QuantumProgramResultSerializationFormat.CUDAQ_SAMPLE_RESULT_JSON_V1,
                serialization=apply_compression[compression_format](
                    sample_serialization
                ),
            )
        except Exception as e:
            raise Exception("unsupport serialization:", compression_format, e)

    def to_cudaq_sample_result(self, **kwargs) -> "cudaq.SampleResult":
        from qio.utils.conversion.program_result.dict_to_cudaq_sample import (
            convert as dict_to_cudaq_sample,
        )

        apply_uncompression = {
            QuantumProgramResultCompressionFormat.NONE: json.loads,
            QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1: zlib_to_dict,
        }

        serialized_sample_result = apply_uncompression[self.compression_format](
            self.serialization
        )

        try:
            apply_unserialization = {
                QuantumProgramResultSerializationFormat.CUDAQ_SAMPLE_RESULT_JSON_V1: dict_to_cudaq_sample,
            }

            return apply_unserialization[self.serialization_format](
                serialized_sample_result
            )
        except Exception as e:
            raise Exception(
                "unsupported unserialization:", self.serialization_format, e
            )

    @classmethod
    def from_qiskit_result(
        cls,
        qiskit_result: "qiskit.result.Result",
        compression_format: QuantumProgramResultCompressionFormat = QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgramResult":
        from qio.utils.conversion.program_result.qiskit_to_dict import (
            convert as qiskit_to_dict_convert,
        )

        qiskit_result_dict = qiskit_to_dict_convert(qiskit_result)

        return cls.from_qiskit_result_dict(qiskit_result_dict, compression_format)

    @classmethod
    def from_qiskit_result_dict(
        cls,
        qiskit_result_dict: Union[str, Dict],
        compression_format: QuantumProgramResultCompressionFormat = QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgramResult":
        if isinstance(qiskit_result_dict, str):
            qiskit_result_dict = json.loads(
                qiskit_result_dict
            )  # Ensure serialization is not ill-formatted

        compression_format = (
            QuantumProgramResultCompressionFormat.NONE
            if compression_format
            == QuantumProgramResultCompressionFormat.UNKNOWN_COMPRESSION_FORMAT
            else compression_format
        )

        apply_compression = {
            QuantumProgramResultCompressionFormat.NONE: json.dumps,
            QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1: dict_to_zlib,
        }

        try:
            return cls(
                compression_format=compression_format,
                serialization_format=QuantumProgramResultSerializationFormat.QISKIT_RESULT_JSON_V1,
                serialization=apply_compression[compression_format](qiskit_result_dict),
            )
        except Exception as e:
            raise Exception("unsupported serialization:", compression_format, e)

    def to_qiskit_result(self, **kwargs) -> "qiskit.result.Result":
        from qio.utils.conversion.program_result.dict_to_qiskit import (
            convert as dict_to_qiskit_convert,
        )
        from qio.utils.conversion.program_result.cirq_to_qiskit import (
            convert as cirq_to_qiskit_convert,
        )
        from qio.utils.conversion.program_result.cudaq_sample_to_qiskit import (
            convert as cudaq_sample_to_qiskit_convert,
        )

        apply_uncompression = {
            QuantumProgramResultCompressionFormat.NONE: json.loads,
            QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1: zlib_to_dict,
        }

        serialization = apply_uncompression[self.compression_format](self.serialization)

        try:
            apply_unserialization = {
                QuantumProgramResultSerializationFormat.QISKIT_RESULT_JSON_V1: dict_to_qiskit_convert,
                QuantumProgramResultSerializationFormat.CUDAQ_SAMPLE_RESULT_JSON_V1: cudaq_sample_to_qiskit_convert,
                QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1: cirq_to_qiskit_convert,
            }

            return apply_unserialization[self.serialization_format](
                serialization, **kwargs
            )

        except Exception as e:
            raise Exception("unsupported serialization:", self.serialization_format, e)

    @classmethod
    def from_cirq_result(
        cls,
        cirq_result: "cirq.Result",
        compression_format: QuantumProgramResultCompressionFormat = QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgramResult":
        from qio.utils.conversion.program_result.cirq_to_dict import (
            convert as cirq_to_dict_convert,
        )

        cirq_result_dict = cirq_to_dict_convert(cirq_result)

        return cls.from_cirq_result_dict(cirq_result_dict, compression_format)

    @classmethod
    def from_cirq_result_dict(
        cls,
        cirq_result_dict: Union[str, Dict],
        compression_format: QuantumProgramResultCompressionFormat = QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1,
    ) -> "QuantumProgramResult":
        if isinstance(cirq_result_dict, str):
            cirq_result_dict = json.loads(
                cirq_result_dict
            )  # Ensure serialization is not ill-formatted

        compression_format = (
            QuantumProgramResultCompressionFormat.NONE
            if compression_format
            == QuantumProgramResultCompressionFormat.UNKNOWN_COMPRESSION_FORMAT
            else compression_format
        )

        apply_compression = {
            QuantumProgramResultCompressionFormat.NONE: json.dumps,
            QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1: dict_to_zlib,
        }

        try:
            serialization = apply_compression[compression_format](cirq_result_dict)

            return cls(
                compression_format=compression_format,
                serialization_format=QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1,
                serialization=serialization,
            )
        except Exception as e:
            raise Exception("unsupported serialization:", e)

    def to_cirq_result(self, **kwargs) -> "cirq.Result":
        from qio.utils.conversion.program_result.dict_to_cirq import (
            convert as dict_to_cirq_convert,
        )
        from qio.utils.conversion.program_result.qiskit_to_cirq import (
            convert as qiskit_to_cirq_convert,
        )

        apply_uncompression = {
            QuantumProgramResultCompressionFormat.NONE: json.loads,
            QuantumProgramResultCompressionFormat.ZLIB_BASE64_V1: zlib_to_dict,
        }

        result_dict = apply_uncompression[self.compression_format](self.serialization)

        try:

            apply_unserialization = {
                QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1: dict_to_cirq_convert,
                QuantumProgramResultSerializationFormat.QISKIT_RESULT_JSON_V1: qiskit_to_cirq_convert,
            }

            return apply_unserialization[self.serialization_format](
                result_dict, **kwargs
            )

        except Exception as e:
            raise Exception(
                "unsupported unserialization:", self.serialization_format, e
            )
